"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.requireWithFakeGlobalScope = exports.requireWithCache = exports.createRequire = exports.requireDependencies = void 0;
const fs_1 = require("fs");
const vm_1 = require("vm");
const path_1 = require("path");
const module_1 = __importDefault(require("module"));
/**
 * Allows to require a series of dependencies provided by their path
 * into a provided module context. It fills and accepts a require
 * cache to ensure each module is loaded once.
 */
function requireDependencies(params) {
    const { context, requireCache, dependencies } = params;
    const requireFn = createRequire(context, requireCache);
    for (const { path, mapExports } of dependencies) {
        const mod = requireFn(path, path);
        for (const mapKey of Object.keys(mapExports)) {
            context[mapExports[mapKey]] = mod[mapKey];
        }
    }
}
exports.requireDependencies = requireDependencies;
function createRequire(context, cache, references, scopedContext = {}) {
    return function requireFn(referrer, specifier) {
        const resolved = require.resolve(specifier, {
            paths: [(0, path_1.dirname)(referrer)],
        });
        const cached = cache.get(specifier) || cache.get(resolved);
        if (cached !== undefined && cached !== null) {
            return cached.exports;
        }
        const module = {
            exports: {},
            loaded: false,
            id: resolved,
        };
        cache.set(resolved, module);
        references === null || references === void 0 ? void 0 : references.add(resolved);
        const fn = (0, vm_1.runInContext)(`(function(module,exports,require,__dirname,__filename,${Object.keys(scopedContext).join(',')}) {${(0, fs_1.readFileSync)(resolved, 'utf-8')}\n})`, context);
        try {
            fn(module, module.exports, requireFn.bind(null, resolved), (0, path_1.dirname)(resolved), resolved, ...Object.values(scopedContext));
        }
        catch (error) {
            cache.delete(resolved);
            throw error;
        }
        module.loaded = true;
        return module.exports;
    };
}
exports.createRequire = createRequire;
function requireWithCache(params) {
    var _a;
    return createRequire(params.context, (_a = params.cache) !== null && _a !== void 0 ? _a : new Map(), params.references, params.scopedContext).call(null, params.path, params.path);
}
exports.requireWithCache = requireWithCache;
function requireWithFakeGlobalScope(params) {
    const resolved = require.resolve(params.path);
    const getModuleCode = `(function(module,exports,require,__dirname,__filename,globalThis,${Object.keys(params.scopedContext).join(',')}) {${(0, fs_1.readFileSync)(resolved, 'utf-8')}\n})`;
    const module = {
        exports: {},
        loaded: false,
        id: resolved,
    };
    const moduleRequire = (module_1.default.createRequire || module_1.default.createRequireFromPath)(resolved);
    function throwingRequire(path) {
        var _a;
        if (path.startsWith('./')) {
            const moduleName = path.replace(/^\.\//, '');
            if (!((_a = params.cache) === null || _a === void 0 ? void 0 : _a.has(moduleName))) {
                throw new Error(`Cannot find module '${moduleName}'`);
            }
            return params.cache.get(moduleName).exports;
        }
        return moduleRequire(path);
    }
    throwingRequire.resolve = moduleRequire.resolve.bind(moduleRequire);
    eval(getModuleCode)(module, module.exports, throwingRequire, (0, path_1.dirname)(resolved), resolved, params.context, ...Object.values(params.scopedContext));
    return module.exports;
}
exports.requireWithFakeGlobalScope = requireWithFakeGlobalScope;
//# sourceMappingURL=require.js.map